# Test program for GLADOS in Python. User supplies a cube, a pixel position and velocity range. 
# The program returns the velocity of the peak S/N value.


import math
import numpy
try:
	import pyfits
except:
	from astropy.io import fits as pyfits
import os
import matplotlib
from matplotlib import pyplot
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
from matplotlib.widgets import Button
import shutil
import sys
from sys import platform
import glob
import multiprocessing
import urllib
import PIL
from PIL import Image, ImageEnhance, ImageFont, ImageDraw
import webbrowser


# Load settings from the file that's always created when GLADoS is called
settingsfile = open('ActualSettingsUsed.txt','r')

# File will always have the same format, so we can just proceed line by line
infile = settingsfile.readline().splitlines()[0].split('=')[1]
dupol = int(settingsfile.readline().splitlines()[0].split('=')[1])
pafile = settingsfile.readline().splitlines()[0].split('=')[1]
pbfile = settingsfile.readline().splitlines()[0].split('=')[1]
mode = settingsfile.readline().splitlines()[0].split('=')[1]
stype = settingsfile.readline().splitlines()[0].split('=')[1]
SNt = float(settingsfile.readline().splitlines()[0].split('=')[1])
SNtp = float(settingsfile.readline().splitlines()[0].split('=')[1])
VWt = float(settingsfile.readline().splitlines()[0].split('=')[1])
VWtp = float(settingsfile.readline().splitlines()[0].split('=')[1])
Vtol = float(settingsfile.readline().splitlines()[0].split('=')[1])
style = settingsfile.readline().splitlines()[0].split('=')[1]		
lowv = settingsfile.readline().splitlines()[0].split('=')[1]		
highv = settingsfile.readline().splitlines()[0].split('=')[1]
adops = int(settingsfile.readline().splitlines()[0].split('=')[1])
# User may or may not have full starjava installed. Could be using a standalone .jar file,
# so need to check for that.	
starline = settingsfile.readline().splitlines()[0]
stardir = starline.split('=')[1].split('stilts')[0]
starfile= starline.split('=')[1].split(stardir)[1]
order = int(settingsfile.readline().splitlines()[0].split('=')[1])		
clip = float(settingsfile.readline().splitlines()[0].split('=')[1])	
niter = int(settingsfile.readline().splitlines()[0].split('=')[1])		
spx = str(settingsfile.readline().splitlines()[0].split('=')[1])
nc = str(settingsfile.readline().splitlines()[0].split('=')[1])		
gs = str(settingsfile.readline().splitlines()[0].split('=')[1])

settingsfile.close()

try:
	if platform == 'win32':
		os.system('cls')
	else:
		os.system('clear')
finally:
	pass


# SANITY CHECKS
if mode <> 'Search' and mode <> 'Review' and mode <> 'Resume' and mode <> 'Complete':
	print 'INVALID MODE SELECTED - ABORTING !'
	print 'Valid modes are search, complete, review and resume.'
	print 'Note that modes are case sensitive !'
	sys.exit()
	
if mode == 'Resume' and (stype <> 'All' and stype <> 'Rejected' and stype <> 'Uncertain' and stype <> 'Accepted'):
	print 'INVALID SOURCE TYPE SELECTED - ABORTING !'
	print 'In resume mode, valid source types are All, Accepted, Uncertain and'
	print 'Rejected.'
	print 'Note that source type is case sensitive !'
	sys.exit()

# Check if the FITS file exists, as long as we're searching the file. If we're reviewing existing
# catalogues then the FITS file isn't needed.	
if not os.path.exists(infile) and (mode == 'Search' or mode == 'Complete'):
	print 'FITS file not found, aborting !'
	sys.exit()
	
if clip <= 0.0:
	print 'Clip value cannot be negative, aborting !'
	sys.exit()
	
if niter < 1:
	print 'No. iterations cannot be less than 1, aborting !'
	sys.exit()
	
if order < 0:
	print 'Polynomial order must be positive, aborting !'
	sys.exit()
	
if SNt <= 1.0:
	print 'A S/N threshold < 1.0 will clip the noise, aborting !'
	sys.exit()
	
if VWt <= 0.0:
	print 'Velocity width threshold must be positive, aborting !'
	sys.exit()
	
if style <> 'Beam' and style <> 'Simple':
	print 'Pixel weighting method invalid, aborting !'
	print 'Method must either be Beam or Simple (case sensitive).'
	sys.exit()
	
if not os.path.exists(stardir+starfile):
	print 'Starlink merging command not found, aborting !'
	sys.exit()



# Changes to the directory where the script is located
abspath = os.path.abspath(__file__)
dname = os.path.dirname(abspath)
os.chdir(dname)


FitsFile = pyfits.open(infile)

image = FitsFile[0].data
header =FitsFile[0].header

sizez = image.shape[0]
sizey = image.shape[1]
sizex = image.shape[2]


# Here get and set parameters we'll need for converting channel numbers to velocity values
# From FITS header :
zref = header['CRPIX3']		# Reference channel number
zvel = header['CRVAL3']		# Velocity of reference channel
zdel = header['CDELT3']		# Velocity width of reference channel

# More parameters - these are easier to set directly
lamda0 = 0.2110611405413	# Should really get this from the file !!
f0     = 1420.40575177E6
c      = 299792458.0
lamref  = ((zvel*lamda0)/c) + lamda0     # Wavelength of reference channel
fref = c/lamref								  # Frequency of reference channel
fplus1 = c /(( ( (zvel+zdel)*lamda0)/c) + lamda0)
deltaf = (fplus1 - fref)


# Assumes that headers, grids are identical to average file
if dupol == 1:
	FitsFileA = pyfits.open(pafile)
	imageA = FitsFileA[0].data

	FitsFileB = pyfits.open(pbfile)
	imageB = FitsFileB[0].data



# Set channel range to search, if needed. Default values search the whole cube.
# Should eventually replace this with velocity range, which is more convenient, and could be specified
# in the main user-parameter section.
if (mode == 'Search') or (mode == 'Complete'):
	if lowv == 'Vmin':
		zmax = sizez
	if highv == 'Vmax':
		zmin = 0
	
	
# If the user has specified velocity values, convert them to channel numbers here
def channelfromvel(velocity):
	k = (c*zref*deltaf) - (c*fref) + (c*f0) + (velocity*zref*deltaf) - (velocity*fref)
	cn= k / (deltaf*(velocity+c))
	return int(cn)

	
if lowv <> 'Vmin':
	zmax = channelfromvel(float(lowv)*1000.0)
	
if highv <> 'Vmax':
	zmin = channelfromvel(float(highv)*1000.0)

	

# Parameters for spatial WCS
crpix1 = header['CRPIX1']
crval1 = header['CRVAL1']
cdelt1 = header['CDELT1']
crpix2 = header['CRPIX2']
crval2 = header['CRVAL2']
cdelt2 = header['CDELT2']


# Beam size, used for weighting the surrounding pixels when extracting spectra
beams = None
beams2= None
# Try and get at least one keyword from the FITS header. We only really need one value.
try:
	beams = header['BMAJ']
except:
	pass
try:
	beams2= header['BMIN']
except:
	pass
	
# Repeat for other possible keywords. Note that keywords are NOT case sensitive.
try:
	beams = header['bmajor']
except:
	pass
try:
	beams2= header['bminor']
except:
	pass
	
try:
	beams = header['beammaj']
except:
	pass
try:
	beams2= header['beammin']
except:
	pass
	
try:
	beams = header['beammajor']
except:
	pass
try:
	beams2= header['beamminor']
except:
	pass
	
#If there's nothing in the keywords, there might be something in the header
bmajors = ['bmaj','bmajor','beammaj','beammajor']
bminors = ['bmin','bminor','beammin','beamminor']
bmjindx = None
bmnindx = None

# Go through the entire header searching for references to any value of BMAJ. If a match is found, it
# will likely be in the "history" section. Extracts the line as an array of space-separated strings.
if beams == None:
	print "Couldn't find the beam size in the FITS header. Searching the history..."
	
	for thing in header[:]:
		headerstring = str(thing)
		for bstring in bmajors:	
			# Convert both strings to lower case to make them case insensitve
			if bstring.lower() in headerstring.lower():
				# Get an array of the history line so we can easily extract the correct value, regardless of
				# the exact line structure
				headlist =headerstring.split()
				for i in range(len(headlist)):
					if bstring.lower() in headlist[i].lower():
						bmjindx = i
					
		# Same procedure for minor axis
		for bstring in bminors:	
			if bstring.lower() in headerstring.lower():
				# Get an array of the history line so we can easily extract the correct value, regardless of
				# the exact line structure
				headlist =headerstring.split()
				for i in range(len(headlist)):
					if bstring.lower() in headlist[i].lower():
						bmnindx = i

# If a match has been found, the numerical value will be in the next index (unless that's used for the
# equals sign).					
if bmjindx <> None:
	if '=' <> headlist[bmjindx+1]:
		try:
			beams = float(headlist[bmjindx+1])
		except:
			pass
	if '=' == headlist[bmjindx+1]:
		try:
			beams = float(headlist[bmjindx+2])
		except:
			pass
	
if bmnindx <> None:
	if '=' <> headlist[bmnindx+1]:
		try:
			beams2 = float(headlist[bmnindx+1])
		except:
			pass
	if '=' == headlist[bmnindx+1]:
		try:
			beams2 = float(headlist[bmnindx+2])
		except:
			pass

# We only need the general beam size. Set this to be the average of the two values if they're both found
if beams <> None and beams2 <> None:
	beams = (beams+beams2)/2.0

# If only the minor axis is set (weird but possible), use that	
if beams == None and beams2 <> None:
	beams = beams2

# If only the major axis is set then there's no need to do anything else because we only use the "beams"
# parameter and that's already been set

	
if beams == None:
	print "ERROR ! Couldn't find the beam size anywhere in the FITS header."
	print "Please edit the header to include the BMAJ keyword and value."
	print "The beam size is used for integrating the spectra ang generating"
	print "various images, so it's quite important."
	sys.exit()



# Assume that cubes are in equatorial sin sky projection format. Keep these as parameters in case we
# later decide to generalise (see code in FRELLED).
wcf = 15.0
prj = 1.0


ncandidates = 0


# FILE/DIRECTORY CLEANING MODULES
# Remove existing sub-directories containing all spectra
def cleansourcedirs():
	if os.path.exists('All'):
		shutil.rmtree('All')
	if os.path.exists('Spectra'):
		shutil.rmtree('Spectra')	

# Remove the subdirectories for the subcategories
def cleansubdirs():
	if os.path.exists('Accepted'):
		shutil.rmtree('Accepted')	
	if os.path.exists('Uncertain'):
		shutil.rmtree('Uncertain')
	if os.path.exists('Rejected'):
		shutil.rmtree('Rejected')

# Remove existing base catalogue files
def cleanbasiccatalogues():	
	if os.path.exists('Catalogue.txt'):
		os.remove('Catalogue.txt')
	if os.path.exists('FullSearchResults.txt'):
		os.remove('FullSearchResults.txt')		
		
# Remove existing sub-category catalogue files
def cleansubcatalogues():
	if os.path.exists('AcceptedSources.txt'):
		os.remove('AcceptedSources.txt')
	if os.path.exists('AcceptedSources.txt'):
		os.remove('AcceptedSources.txt')
	if os.path.exists('UncertainSources.txt'):
		os.remove('UncertainSources.txt')			
	if os.path.exists('RejectedSources.txt'):
		os.remove('RejectedSources.txt')		
		
# Remove existing directory holding SDSS images
def cleanimagedir():
	if os.path.exists('SDSS'):
		shutil.rmtree('SDSS')
		
	if os.path.exists('FluxMaps'):
		shutil.rmtree('FluxMaps')
		
	if os.path.exists('PVMaps'):
		shutil.rmtree('PVMaps')
		
	if os.path.exists('GMaps'):
		shutil.rmtree('GMaps')	


# Check if the source catalogues and spectra in directories agree. Useful for "resume" mode, where
# it's important that everything is kept consistent.
def sourcecheck():
	for sourcetype in ['Accepted','Rejected','Uncertain','All']:
		scname = sourcetype+'Sources.txt'
		subcatdir  = dname+'/'+str(sourcetype)+'/'
		
		# Exception for using the base catalogue
		if sourcetype == 'All':
			scname = 'Catalogue.txt'
			subcatdir = dname+'/All/'
			
		subcatfile = open(scname,'r')
		subcatfile.readline()	
		
		# First check that all sources in each file are found in the corresponding directory 
		for line in subcatfile:
			if scname <> 'Catalogue.txt':
				name, x, y, z, ra, dec, vel, sn, vw, rms = line.split()
			if scname == 'Catalogue.txt':
				name, x, y, z, ra, dec, vel, sn, vw, rms, gid, gs = line.split()
			imname = str(name)+'.png'
			if not os.path.exists(subcatdir+str(imname)):
				print 'ERROR !',imname,'not found in',subcatdir,'directory'
				print 'Halting code until you sort out this horrible mess.'
				sys.exit()
				
		subcatfile.seek(0)
		subcatfile.readline()
		status = False
		
		# Now do the reverse and check that all sources found in the directory are also listed in 
		#the appropriate catalogue file.
		os.chdir(subcatdir)
		files = glob.glob('*.png')	
		for filename in files:
			sourcename = filename.split('.png')[0]
			for line in subcatfile:
				if scname <> 'Catalogue.txt':
					name, x, y, z, ra, dec, vel, sn, vw, rms = line.split()
				if scname == 'Catalogue.txt':
					name, x, y, z, ra, dec, vel, sn, vw, rms, gid, gs = line.split()
				name = str(name)
				if sourcename == name:
					status = True
					
			if status == False:
				print 'ERROR !',sourcename,'not found in',subcatfile,'file'
				print 'Halting code until you sort out this horrible mess.'
				sys.exit()
		
		os.chdir(dname)		
		subcatfile.close()
		


# Convert spatial pixesl to WCS
def wcspix(x,y):
	dec = ((y - crpix2)*cdelt2) + crval2
	ra = (crval1/wcf) - -cdelt1*(x - crpix1)/(wcf)*(1.0/math.cos(math.radians(prj*dec)))
	
	# Now convert to sexigesimal format. First for declination :
	dd = int(math.modf(dec)[1])
	dm = int(math.modf(math.modf(dec)[0]*60.0)[1])
	ds = math.modf(math.modf((math.modf(dec)[0]*60.0))[0])[0]*60.0
	
	# Prevent rounding errors from displaying 60 seconds or minutes of dec and also
	# check for negative declination cases
	if dec < 0.0:
		dm = abs(dm)
		ds = abs(ds)
		
	if (dec > 0.0) and (abs(float(ds) - 60.0) <=0.1):
		ds = 0
		dm = dm + 1
	if (dec > 0.0) and (abs(float(dm) - 60.0) <=0.01):
		dm = 0
		dd = dd + 1
	
	if (dec < 0.0) and (abs(float(ds) - 60.0) <=0.1):
		ds = 0
		dm = dm - 1	
	if (dec < 0.0) and (abs(float(dm) - 60.0) <=0.01):
		dm = 0
		dd = dd - 1
		
	# str().zfill(2) just ensures have trailing zeros for nicer formatting 
	# %2.f ensure only 2 decimal places

	sexd = str(dd).zfill(2)+str(':')+str(dm).zfill(2)+str(':')+str("%2.f" % ds)

	# Remove any spaces. Even though the formatting is the same, for some reason decs automatically 
	# rounds to 0 d.p., whereas ras round to 2 d.p. So this is only really needed for dec.
	dec = sexd.replace(' ','')
	
	
	# Now for right ascension
	rh = int(math.modf(ra)[1])
	rm = int(math.modf(math.modf(ra)[0]*60.0)[1])
	rs = math.modf(math.modf((math.modf(ra)[0]*60.0))[0])[0]*60.0
	
	# Prevent rounding errors from displaying 60 seconds or minutes of time
	if (abs(float(rs) - 60.0) <= 0.01):
		rs = 0
		rm = rm + 1
	if (abs(float(rm)-60.0) <= 0.01):
		rm = 0
		rh = rh + 1	
		
	sexr = str(rh).zfill(2)+str(':')+str(rm).zfill(2)+str(':')+str("%.2f" % rs)
	
	if ra < 0.0:
		sexr = str(rh).zfill(2)+str(':')+str(abs(rm)).zfill(2)+str(':')+str("%.2f" % abs(rs))
	
	sexr = sexr.replace(' ','')
	
	ra = sexr
		
	return ra, dec



# Convert channel number to WCS velocity
def velocity(z):
	vel = -(c*( (((z - zref)*deltaf) + fref - f0)/(((z - zref)*deltaf) + fref) ))
	return vel/1000.0



# Extract a spectrum at the specified location, optionally weighting by the beam as in mbspect.
# Also adds a column to hold the velocity values
def extractspec(xp,yp,zmin,zmax,style,pol):

	# Create an array to hold the channel numbers
	channel = numpy.array([i for i in range(zmin,zmax)])
	if pol == 'Average':
		flux = numpy.array(image[zmin:zmax,yp,xp])
	if pol == 'A':
		flux = numpy.array(imageA[zmin:zmax,yp,xp])
	if pol == 'B':
		flux = numpy.array(imageB[zmin:zmax,yp,xp])
	
			
	# Check if surrounding spectra are good. If they're not, don't proceed
	ngood = 0
	if style == 'Beam':
		for i in [xp-1,xp,xp+1]:
			for j in [yp-1,yp,yp+1]:
				# If any problem with the specturm, set it to NaN so it will be counted as bad
				try:
					if pol == 'Average':
						newspec = image[zmin:zmax,j,i]
					if pol == 'A':
						newspec = imageA[zmin:zmax,j,i]
					if pol == 'B':
						newspec = imageB[zmin:zmax,j,i]	
				except:
					newspec = [numpy.nan]
					
				if numpy.isnan(newspec).any() == False:
					ngood = ngood +1
				
	# If mbspect-style integration is requested, and it's possble, do the integration before
	# doing the baseline fitting.
	if style == 'Beam' and ngood == 9:
		SW = 1.0
		
		for i in [xp-1,xp,xp+1]:
			for j in [yp-1,yp,yp+1]:
				dr = (float(i)-float(xp))*cdelt1
				dd = (float(j)-float(yp))*cdelt2
				W  = math.exp(-2.0*math.log10(4.0)*(dr**2.0 +dd**2.0)/(beams*beams))
				if (i <> xp) and (j <> yp):
					if pol == 'Average':
						flux = flux + W*image[zmin:zmax,j,i]
					if pol == 'A':
						flux = flux + W*imageA[zmin:zmax,j,i]
					if pol == 'B':
						flux = flux + W*imageB[zmin:zmax,j,i]	
					SW = SW + W*W
					
		flux = flux / math.sqrt(SW)
		SW = 1.0
	
	
	# Create array of velocity values
	vchans = velocity(channel)
			
	# Combine the channel and flux values into a single 2D array
	spectra = numpy.column_stack((channel,vchans,flux))
	# Note that unforunately this converts the channel values to floats
		
	return spectra



# Measures the rms of a supplied spectrum, using a sigma-clipped polynomial. Optionally can also
# subtract the baseline and add a S/N column.
def spectralfit(spectrum, dosub, fitsn):
	
	# Baseline subtraction with sigma clipping
	
	# Extract arrays of channel numbers (as integers) and flux values
	channel = numpy.array(spectrum[:,0]).astype(int)
	flux = spectrum[:,2]
				
	
	# Problem - channel array contains the abosolute channel numbers. These will agree with the local array
	# indicies as long as the full velocity range is used, but otherwise they will not. We need to provide
	# the local indicies for the polyfitting procedure. Since this function only alters the flux values in the
	# spectrum array, there should be no need to do anything else except alter the tempchan array that's used.
	 		
	tempchan = numpy.array([i for i in range(0,len(channel))])
	
			
	nanlist=[]
	for i in range(0,len(flux)):
		if numpy.isnan(flux[i]):
			nanlist.append(i)
		
	tempflux = numpy.delete(flux,nanlist)
	tempchan = numpy.delete(tempchan,nanlist)
	
		
	# Need to account for the case where there weren't enough values to proceed. Although it would be
	# better to skip these, it's easier just to replace them with blanks.
	if len(tempflux)<=order:
		flux[:]=0.0
		tempflux = flux
		tempchan = numpy.array([i for i in range(0,len(channel))])
		
		
	if len(tempflux)>order :
		for i in range(0,niter):
			length = len(tempchan)
					
			# Fit a polynomial
			temppoly = numpy.polyfit(tempchan, tempflux,order)
			tempmodl = numpy.array([0.0 for k in range(0,length)])
				
			for f in range(0,order+1):
				fn = float(-f + order)
				tempmodl = tempmodl + temppoly[f]*pow(tempchan,fn)
					
			tempspec = tempflux - tempmodl
				
			rms = numpy.std(tempflux)
			av  = numpy.median(tempflux)
			
			# Find the location of values to remove
			deviants = numpy.where(abs(tempflux - av) > clip*rms)
				
			if (len(deviants)>0 and len(deviants) < length):
				tempflux = numpy.delete(tempflux,deviants)
				tempchan = numpy.delete(tempchan,deviants)
		
				
		# Fit a polynomial to the original spectrum, masking all deviant values		
		poly = numpy.polyfit(tempchan,flux[tempchan],order)
		model = numpy.array( [0.0 for k in range(0,len(flux))])
		for f in range(0,order+1):
			fn = float(-f + order)
			model = model + poly[f]*pow(channel,fn)
			
		fitted = flux - model
		
		# This will only be returned if dosub is true, so the subtracted flux column may as well be set here.
		spectrum[:,2] = fitted
				
		# Spectrum in S/N format
		if fitsn == True:	
			snspec = fitted / rms
			spectrum = numpy.column_stack((spectrum,snspec))
		
		if dosub == True:
			return spectrum, rms
			
		if dosub == False:
			return rms
			
		# User might ask for several things :
		# dosub = True
		# - Returns the baseline-corrected spectrum and rms
		# dosub = True; fitssn = True
		# - Returns the baseline-corrected spectrum with an extra S/N colum, and the rms
		# dosub = False
		# - Returns only the rms based on the sigma-clipped polynomial but leaving the original spectrum
		# unchanged. 
		


# Find where the peak values are in a supplied spectrum
def findpeakindex(spectrum):
	# Interesting quirk. This works, even though this argmax is from numpy :
	#maxindex = spectrum.argmax(axis=0)
	# But to deal with the case of masked channels full of nans, we need to use numpy.nanargmax.
	# This will return an array of values giving the index at which each colum has its greatest value.
	# We can therefore get the channel number of the greatest flux from this final value
	# Error trapping to deal with cases of pure NaNs.
	try:
		maxindex = numpy.nanargmax(spectrum,axis=0)
	except:
		maxindex = [-999, -999, -999, -999]
	
	return maxindex



# Find the velocity of a specified channel in a spectrum
def velchan(spectrum, channel):
	velchan = spectrum[channel,1]
	return velchan
	


# Find the velocity width of a peak in a supplied spectrum
def vwidth(spectrum,start):
	# Velocity of peak index
	vpeak = spectrum[start,1]
	
	# Index of channel to start the search
	channel = start
	
	# Starting flux value
	pflux = spectrum[start,2]
	flux = pflux
	
	
	# Go forward through the velocity indicies to find the first channel where the flux is below 50%
	# of the peak value.
	# First go forward in the channels to get the lowest velocity :
	while flux > (pflux/2.0):
		channel = channel + 1
		
		# Usually this will exit after a few channels since most detections are not that wide. But we
		# need to keep going to measure the "true" width of the source. We have to force the loop to
		# end if the velocity reaches the limits of the spectrum.
		# Might expect to do :
		##if (channel >= zmax-1):
		# But we can't do this. "Channel" is really local index, not actual channel number !
		if (channel >= len(spectrum)-1):
			#channel = zmax-1
			channel = len(spectrum)-1
			break
		
		flux = spectrum[channel,2]
		
		
	# Use the last channel value found for the velocity width measurement later
	vminchan = channel
	
	
	# Now repeat but backwards to get the minimum channel value
	channel = start
	flux = pflux
	
	
	while flux > (pflux/2.0):
		channel = channel - 1
		
		if (channel <= 0):
			channel = 0
			break
		
		flux = spectrum[channel,2]
	
		
	# Set the end channel for the velocity width measurement
	vmaxchan = channel
	
	
	# We now have the channel indicies for the point just after the intensity intersects the 50%
	# lines. To find the intersection point we need the actual velocity values just before and
	# after.
	vmin  = spectrum[vminchan,1]
	vmin1 = spectrum[vminchan-1,1]
	fmin  = spectrum[vminchan,2]
	fmin1 = spectrum[vminchan-1,2]
	

	m = (fmin1 - fmin) / (vmin1 - vmin)
	c = fmin - (m*vmin)
	
	# Minimum velocity intersection
	vini = ((pflux/2.0) - c) / m
	
	vmax = spectrum[vmaxchan,1]
	vmax1= spectrum[vmaxchan+1,1]
	fmax = spectrum[vmaxchan,2]
	fmax1= spectrum[vmaxchan+1,2]
	
	
	m = (fmax1 - fmax) / (vmax1 - vmax)
	c = fmax - (m*vmax)
	
	# Maximum velocity intersection
	vfin = ((pflux/2.0) - c) / m
	
	return vfin-vini	



# Peak index array returns -999 if the array is of pure nans. If so, set the S/N to be below the
# threshold value so it will be rejected.
def searchspec(spectrum,xp,yp,rms):	
	global ncandidates
	peakchan = findpeakindex(spectrum)[3]
	if peakchan >=0:
		peaksn = spectrum[peakchan,3]
	elif peakchan == -999:
		peaksn = SNt - 1.0
	return peaksn, peakchan
	

# CHELL : Clears HI Excessive Line Locations
# Basically this bit does the merging of adjacent pixels.
def CHELL():
	# If the user has starjava installed, there will be binary file for the STILTS command.
	# If not, they'll be using a .jar file
	# Wrap the command in DOUBLE quotes as on Windows spaces in directory names can cause
	# problems
		
	# Have to enclose directory path in double quotes on Windows, not single quotes !
	if '.jar' not in starfile:
		prefix = '\"'+stardir+starfile+'\"'
	if '.jar' in starfile:
		prefix = 'java -jar '+'\"'+stardir+starfile+'\"'
	
	# USE OF SINGLE AND DOUBLE QUOTES HERE IS CRITICAL FOR WINDOWS-LINUX CROSS-COMPATABILITY !!!
	print 'Creating merged catalogue...'	
	os.system(prefix+' tmatch1 ifmt=ascii in=FullSearchResults.txt ofmt=ascii out=Temp1.txt matcher=3d_anisotropic params=\"'+spx+' '+spx+' '+nc+'\" values='+'\"x y z\"')
	
	# Warn the user if nothing found
	if os.path.exists(dname+'/Temp1.txt') == False:
		print 'WARNING ! No sources found !'
	
	# Only proceed if the matched catalogue exists
	if os.path.exists(dname+'/Temp1.txt') == True:	
		os.system(prefix+' tpipe ifmt=ascii in=Temp1.txt cmd=\"sort GroupSize\" ofmt=ascii out=Temp2.txt')
		os.system(prefix+' tpipe ifmt=ascii in=Temp2.txt cmd=\"select GroupSize>='+gs+'\" ofmt=ascii out=Temp3.txt')
		os.system(prefix+' tpipe ifmt=ascii in=Temp3.txt cmd=\"sort -down SN\" ofmt=ascii out=Temp4.txt')
		os.system(prefix+' tmatch1 ifmt=ascii ofmt=ascii in=Temp4.txt out=Catalogue.txt action=keep1 matcher=exact values=GroupID')
				
		os.remove('Temp1.txt')
		os.remove('Temp2.txt')
		os.remove('Temp3.txt')
		os.remove('Temp4.txt')
	
		infile = open('Catalogue.txt','r')
		nlines = len(infile.readlines())
		infile.close()
	
		print nlines-1,'candidates after merging.'
	


# WHEATLEY : When HI Evades Analysis, This Lists EverYthing
# Plots the spectra of all sources from the final catalogue file
def WHEATLEY():
	print 'Plotting spectra of merged catalogue...'

	# Create directory to hold full source maps, combining spectra and RBG images and whatnot
	if not os.path.exists('All'):
		os.mkdir('All')
	
	# Create directory to hold spectra images	
	if not os.path.exists('Spectra'):
		os.mkdir('Spectra')	
	
	# Create directory to hold SDSS images	
	if not os.path.exists('SDSS'):
		os.mkdir('SDSS')
		
	# Create directory to hold moment maps	
	if not os.path.exists('FluxMaps'):
		os.mkdir('FluxMaps')
		
	if not os.path.exists('PVMaps'):
		os.mkdir('PVMaps')
		
	if not os.path.exists('GMaps'):
		os.mkdir('GMaps')		
	

	# LaTeX like fonts in the figure
	plt.rc('font', family='serif')

	titlessize = 15
	numbersize = 12

	#csfont = {'fontname':'Liberation Serif','fontsize':titlessize}
	#csfont = {'fontname':'Times','fontsize':titlessize}

	from matplotlib.ticker import AutoMinorLocator
	XminorLocator = AutoMinorLocator()
	YminorLocator = AutoMinorLocator()

	# Disable interactive plotting
	pyplot.ioff()

	# Load the catalogue file
	catfile = open('Catalogue.txt','r')
	catfile.readline()

	# Get the parmaters we'll need for the moment and PV map. Some of these are fixed and calculated
	# here but others are source-dependent and calculated in the loop below.
	# Moment map size :
	mmapsze = abs(10.0*beams/cdelt1)		# Assuming that cdelt1 and 2 are the same
	# Round up to nearest 5 pixels
	mmapsze = int(int(math.ceil(mmapsze/5.0))*5.0)
	
	# Get the median rms. PROBABLY NOT NEEDED !
	rmsvals =[]
	for line in catfile:
		name, x, y, z, ra, dec, vel, SN, vw, rms, gid, gs = line.split()
		rmsvals.append(float(rms))
	
	rmsvals = numpy.asarray(rmsvals)
	medrms = numpy.median(rmsvals)
	
	# Reset the file
	catfile.seek(0)
	catfile.readline()


	for line in catfile:
		name, x, y, z, ra, dec, vel, SN, vw, rms, gid, gs = line.split()

		# Get basic properties, some from file, some from spectrum
		outname = str(name)+'.png'
		x = int(x)
		y = int(y)
		SN = float(SN)
		vw = float(vw)
		vel = float(vel)
		rms = float(rms)
	
		# Since we don't alter the cube itself in the above procedures, we need to re-extract the spectra
		# First get the average of the two polarisations
		# Give the spectrum a name prefix but no need to do this for the other parameters since we don't
		# need them here
		avspectrum = extractspec(x,y,zmin,zmax,style,'Average')
	
		avspectrum[:,2] = avspectrum[:,2]*1000.0
	
		# Get S/N
		avspectrum, rms = spectralfit(avspectrum, dosub=True, fitsn=True)
	
		# Get velocities
		vels = avspectrum[:,1]
	
		# Get the flux. Modify for plotting by using values in mJy
		flux = avspectrum[:,2]#*1000.0
	
		# Get the S/N level
		sn = avspectrum[:,3] 
			
		# Get the peak flux value
		peakchan = findpeakindex(avspectrum)[2]
		pf = flux[peakchan]
		
		# Get the velocity range to plot. Won't matter if no data values here, will just show as blank
		vmin = vel - 1000.0
		vmax = vel + 1000.0
		
		# Set the flux level to display, based on the rms and peak value
		fmin = -5.0*float(rms)#*1000.0
		fmax = 1.25*float(pf)
	
		# Get S/N limits
		snmin = fmin / rms
		snmax = fmax /rms
		
		
		# Now get and correct the spectra for the individual polarisations, if necessary
		if dupol == 1:
			Aspectrum = extractspec(x,y,zmin,zmax,style,'A')
			Aspectrum[:,2] = Aspectrum[:,2]*1000.0
			Aspectrum, dummy = spectralfit(Aspectrum, dosub=True, fitsn=True)
			Avels = Aspectrum[:,1]
			Aflux = Aspectrum[:,2]#*1000.0
		
			Bspectrum = extractspec(x,y,zmin,zmax,style,'B')
			Bspectrum[:,2] = Bspectrum[:,2]*1000.0
			Bspectrum, dummy = spectralfit(Bspectrum, dosub=True, fitsn=True)
			Bvels = Bspectrum[:,1]
			Bflux = Bspectrum[:,2]#*1000.0
	
	
		
		# Plot the spectra
		fig, ax1 = plt.subplots()
		fig.set_size_inches(750.0/100.0,750.0/100.0)
		
		# Plot individual polarisations, if selected
		if dupol == 1:
			ax1.plot(Avels, Aflux, 'b-', color='red', linewidth=0.25)
			ax1.plot(Bvels, Bflux, 'b-', color='blue', linewidth=0.25)
		
		ax1.plot(vels, flux, 'b-', color='black', linewidth=0.5)
	
		pyplot.title(str(name))
	
		ax1.set_xlim(vmin,vmax)	
	
		ax1.set_xlabel('Velocity / km/s')
	
		mn, mx = ax1.set_ylim(fmin,fmax)
		ax1.set_ylabel('Flux / mJy')
	
		matplotlib.pyplot.text(0.05, 0.95, 'Peak S/N ='+str(round(SN,2)), transform=ax1.transAxes)
		matplotlib.pyplot.text(0.05, 0.90, 'Velocity ='+str(round(vel,2))+' km/s', transform=ax1.transAxes)
		matplotlib.pyplot.text(0.05, 0.85, 'V. width ='+str(round(vw,2))+' km/s',transform=ax1.transAxes)
		
		matplotlib.pyplot.text(0.70, 0.95, 'R.A. ='+str(ra),transform=ax1.transAxes)
		matplotlib.pyplot.text(0.70, 0.9, 'Dec. ='+str(dec),transform=ax1.transAxes)
		
		ax2 = ax1.twinx()
		ax2.set_ylim(snmin, snmax)
		ax2.set_ylabel('S/N')
	
		mn, mx = ax2.set_ylim(snmin,snmax)
	
		XminorLocator = AutoMinorLocator()
		YminorLocator = AutoMinorLocator()
	
		ax1.xaxis.set_minor_locator(XminorLocator)
		ax2.yaxis.set_minor_locator(YminorLocator)
	
	
		YminorLocator = AutoMinorLocator()
	
		ax1.yaxis.set_minor_locator(YminorLocator)
		
		ax1.tick_params(which='major', length=10)
		ax1.tick_params(which='minor', length=5)
	
		ax2.tick_params(which='major', length=10)
		ax2.tick_params(which='minor', length=5)
				
		pyplot.savefig(dname+'/Spectra/'+outname,dpi=100,facecolor='white',bbox_inches='tight')
	
		# If we don't close the figure, it will open them all at once !
		pyplot.close()
		
		
		
		# RETRIEVE SDSS IMAGES
		ra = str(ra)
		dec = str(dec)
		fullra = ra.split(':')
		fulldc = dec.split(':')
		
		rahh = float(fullra[0])
		ramm = float(fullra[1])
		rass = float(fullra[2])
		decra= (rahh + (ramm/60.0) + (rass/3600.0))*15.0
		
		dcdd = float(fulldc[0])
		dcmm = float(fulldc[1])
		dcss = float(fulldc[2])
		decdc= dcdd + (dcmm/60.0) + (dcss/3600.0)
		
		# Set the image size to be equal to the beam size at the SDSS maximum resolution
		imres = 0.4		# SDSS smallest pixel scale in arcseconds
		imsize = int(beams/(imres/3600.0))
		
		# If the image size exceeds the SDSS maximum size, alter the resolution
		if imsize > 2048:
			imsize = 2048
			imres = (beams*3600.0) / float(imsize)	# Pixel scale in arcseconds
		
		urlprefix = 'http://skyserver.sdss.org/dr13/SkyServerWS/ImgCutout/getjpeg?TaskName=Skyserver.Chart.Imagggggge&ra='
				
		url = urlprefix+str(decra)+'&dec='+str(decdc)+'&width='+str(imsize)+'&height='+str(imsize)+'&scale='+str(imres)
		
		imname = str(name)+'_RGB.jpg'
		
		# Just in case something goes wrong with retreiving the image, remove exisitng image first - prevents
		# it from being re-used incorrectly
		try:
			os.remove(dname+'RGBs/'+str(imname))
		except:
			pass
		
		sdssimage = urllib.URLopener()
		sdssimage.retrieve(url,dname+'/SDSS/'+str(imname))
		
		
		
		# CREATE FLUX MAPS
		# Use min, max functions to prevent exceeding the boundaries
		fxmin = max(0, int(float(x) - float(mmapsze)/2.0))
		fxmax = min(sizex, int(float(x) + float(mmapsze)/2.0))
		fymin = max(0, int(float(y) - float(mmapsze)/2.0))
		fymax = min(sizey, int(float(y) + float(mmapsze)/2.0))		
		
		fxrange = fxmax-fxmin
		fyrange = fymax-fymin
		
		cwidth = abs(zdel)/1000.0
				
		cmin = max(0, int(float(z) - float(vw)/cwidth))
		cmax = min(sizez, int(float(z) + float(vw)/cwidth))
		
		# Now we need to work out how to deal with edge cases, i.e. when the x/y size < mmapsze
		# No. channels doesn't matter since it will just integrate over whatever range is specified.
		# We want mmap to have a fixed size. Let the values outside the extracted image remain zero.
		# Then we simply need to add a position offset.
		
		fxoffset = int((float(mmapsze) - float(fxrange))/2.0)
		fyoffset = int((float(mmapsze) - float(fyrange))/2.0)
							
		# Moment map will be of fixed size, though the part of the image we combine with it will vary.
		# Regions that would extend beyond the data will be set to zero and only those pixels inside
		# the requested area will be added.
		mmap = numpy.zeros((mmapsze,mmapsze))
				
		for i in range(int(cmin),int(cmax)):
			# Have to specify exact range of map to alter, otherwise it will truncate the map to the
			# range where the data should be altered
			mmap[fyoffset:(fyoffset+fyrange),fxoffset:(fxoffset+fxrange)] = mmap[fyoffset:(fyoffset+fyrange),fxoffset:(fxoffset+fxrange)]+image[i,fymin:fymax,fxmin:fxmax]

			
		fig = plt.figure()	
		# This works since dpi=100, so images will have correct size and shape
		fig.set_size_inches(float(fxrange)/100.0,float(fyrange)/100.0)
		# Trick to remove any axes and whitespace from the images
		ax = plt.Axes(fig, [0.,0.,1.,1.,])
		ax.set_axis_off()
		fig.add_axes(ax)
		plt.set_cmap('gray')
		
		mmapfile = name+'_Mom0.png'
		
		# Flux values for display range. Should restrict the maximum value to the central area !
		# This isn't perfect but it works reasonably well
		immin = numpy.nanmin(mmap)
		
		# Get the peak flux from are area within the central beam
		beampix = (int(math.ceil(beams/abs(cdelt1))))
		pmin = int((float(fxrange)/2.0)-(beampix/2.0))
		pmax = int((float(fxrange)/2.0)+(beampix/2.0))		
		immax = numpy.nanmax(mmap[pmin:pmax,pmin:pmax])
					
		ax.imshow(mmap,aspect='equal',interpolation='none',vmin=immin,vmax=immax)
		plt.gca().invert_yaxis()
				
		plt.savefig(str(dname)+'/FluxMaps/'+str(mmapfile),dpi=100,facecolor='black')
		plt.close(fig)
		
		
		# CREATE PV MAP
		# Stick with square plots. VW will usually be inaccurate, but will sometimes be reasonable.
		# This would be a lot easier if we had a more reliable velocity estimate...
		pvpx = max(mmapsze, vw/abs(zdel))
		
		xmin = max(0, int(float(x) - float(pvpx)/2.0))
		xmax = min(sizex, int(float(x) + float(pvpx)/2.0))
		chmn = max(0, int(float(z) - float(pvpx)/2.0))
		chmx = min(sizez, int(float(z) + float(pvpx)/2.0))		
		
		xrange = xmax-xmin
		vrange = chmx-chmn
		
		xoffset = int((float(pvpx) - float(xrange))/2.0)
		voffset = int((float(pvpx) - float(vrange))/2.0)
		
		pvmap = numpy.zeros((pvpx,pvpx))
				
		pvmap = pvmap[voffset:(voffset+vrange),xoffset:(xoffset+xrange)]+image[chmn:chmx, y, xmin:xmax]
		
		immin = numpy.nanmin(pvmap)
		immax = numpy.nanmax(pvmap[pmin:pmax,pmin:pmax])
		
		fig = plt.figure()	
		# This works since dpi=100, so images will have correct size and shape
		fig.set_size_inches(float(xrange)/100.0,float(vrange)/100.0)
		# Trick to remove any axes and whitespace from the images
		ax = plt.Axes(fig, [0.,0.,1.,1.,])
		ax.set_axis_off()
		fig.add_axes(ax)
		plt.set_cmap('gray')
				
		pvmapfile = name+'_PV.png'
		
		ax.imshow(pvmap,aspect='equal',interpolation='none',vmin=immin,vmax=immax)
		plt.gca().invert_yaxis()
				
		plt.savefig(str(dname)+'/PVMaps/'+str(pvmapfile),dpi=100,facecolor='black')
		plt.close(fig)
		
		
		# CREATE DETECTION MAP
		# Plots all the sources flagged as possible detections by GLADOS in this region, pre-merging
		dmp = numpy.zeros((mmapsze,mmapsze))
		dmp[:,:] = 'nan'
		ngoodpix = 0
		fullcatfile = open('FullSearchResults.txt','r')
		fullcatfile.readline()
		
		# Go through the full catalogue file and add these to the map if needed.
		# Need to convert from the absolute coordinates in the original file to the positions in the array.
		# Easiest way to do this is to find the absolute value of the boundary pixels of the blank image :
		mapminx = int(float(x) - float(mmapsze)/2.0)
		mapmaxx = int(float(x) + float(mmapsze)/2.0)
		mapminy = int(float(y) - float(mmapsze)/2.0)
		mapmaxy = int(float(y) + float(mmapsze)/2.0)
		
		
		for line in fullcatfile:
			# Have to avoid conflict with existing values
			thisname, thisx, thisy, thisz, thisra, thisdec, thisvel, thissn, thisvw, thisrms = line.split()
			thisx = int(thisx)
			thisy = int(thisy)
			thissn = float(thissn)
			thisvel = float(thisvel)
			
			# fxmin, max truncated to avoid exceeding edges. Not good for trying to assign individual pixels
			# But we already know the current souce (position x, y, z) is within the boundaries, and we know
			# the map size.
			# Therefore absolute value of minx pixel in map = x-mmapsze/2
			# Then to get position of array to adjust we just do thisx - minx
					
			if thisx > mapminx and thisx < mapmaxx and thisy > mapminy and thisy < mapmaxy and abs(thisvel-vel) < Vtol:
				
				# First convert to 0.0 if value is still nan
				if math.isnan(dmp[thisy-mapminy,thisx-mapminx]) == True:
					dmp[thisy-mapminy,thisx-mapminx] = 0.0
					
				# Now add the value
				dmp[thisy-mapminy,thisx-mapminx] = dmp[thisy-mapminy,thisx-mapminx] + thissn
			
				ngoodpix = ngoodpix + 1
				
		fullcatfile.close()
		
		
		fig = plt.figure()	
		# This works since dpi=100, so images will have correct size and shape
		fig.set_size_inches(float(fxrange)/100.0,float(fyrange)/100.0)
		# Trick to remove any axes and whitespace from the images
		ax = plt.Axes(fig, [0.,0.,1.,1.,])
		ax.set_axis_off()
		fig.add_axes(ax)
		plt.set_cmap('rainbow')
		
		gmapfile = name+'_Gmap.png'
		
	
		# Flux values for display range. Should restrict the maximum value to the central area !
		# This isn't perfect but it works reasonably well
		immin = numpy.nanmin(dmp)
		
		# Get the peak flux from are area within the central beam
		immax = numpy.nanmax(dmp)
					
		ax.imshow(dmp,aspect='equal',interpolation='none',vmin=immin,vmax=immax)
		plt.gca().invert_yaxis()
				
		plt.savefig(str(dname)+'/GMaps/'+str(gmapfile),dpi=100,facecolor='white')
		plt.close(fig)
		
		
		# RESIZE AND COMBINE IMAGES
		# The other plot files will be large enough anyway, but it's helpful to resize the PV/flux maps
		# so that they're easier to view externally. We're using the SDSS image as the base, so make
		# the flux map half its height.
		flxfile = dname+'/FluxMaps/'+str(mmapfile)
		rgbfile = dname+'/SDSS/'+str(imname)
		
		# Open the image and get its size. Use floats as we need to find the aspect ratio.
		flximg = PIL.Image.open(flxfile)
		fhsize = float(flximg.size[0])
		fvsize = float(flximg.size[1])
		
		# Same for the SDSS image
		rgbimg = PIL.Image.open(rgbfile)
		rhsize = float(rgbimg.size[0])
		rvsize = float(rgbimg.size[1])
		
				
		# Here it's also helpful to adjust the brightness and contrast of the SDSS image
		bright = ImageEnhance.Brightness(rgbimg)
		enhanced_im = bright.enhance(1.35)
		contr = ImageEnhance.Brightness(enhanced_im)
		enhanced_im = contr.enhance(1.35)
		rgbimg = enhanced_im
		
		# And while we're at it, add some text
		draw = ImageDraw.Draw(rgbimg)	
		#font = ImageFont.truetype('/usr/share/fonts/truetype/ttf-liberation/LiberationSerif-Regular.ttf', 20)
		# This doesn't work on Windows ! Now uses matplotlib to annotate instead (see below) as this is
		# reliable everywhere.
		
		
		# Get the size of the detection map
		dmpimg = PIL.Image.open(dname+'/GMaps/'+gmapfile)
		dhsize = float(dmpimg.size[0])
		dvsize = float(dmpimg.size[1])
		
		
		# We want to resize the flux map to have the same height as half the SDSS image,
		# preserving its aspect ratio
		vscale = (rvsize/fvsize)*0.5
		flximg = flximg.resize((int(fhsize*vscale),int(fvsize*vscale)))		
		# Can overwite this file, no point storing the tiny images
		flximg.save(flxfile)
		
		# Get the new horizontal size, we'll need this for making the final combined image
		fvhsize = float(flximg.size[0])
		
		# Add some text to this too
		draw = ImageDraw.Draw(flximg)
		
			
		# Similar procedure for the PV maps
		pvfile = dname+'/PVMaps/'+str(pvmapfile)
		pvimg = PIL.Image.open(pvfile)
		
		phsize = float(pvimg.size[0])
		pvsize = float(pvimg.size[1])
		
		vscale = (rvsize/pvsize)*0.5
		
		pvimg = pvimg.resize((int(phsize*vscale),int(pvsize*vscale)))
		pvimg.save(pvfile)
		
		phsize = float(pvimg.size[0])
		
		draw = ImageDraw.Draw(pvimg)
		
		
		# Same for detection image
		gvfile = dname+'/GMaps/'+str(gmapfile)
		gvimg = PIL.Image.open(gvfile)
		
		gvhsize = float(gvimg.size[0])
		gvvsize = float(gvimg.size[1])
		
		vscale = (rvsize/gvvsize)*0.5
		
		gvimg = gvimg.resize((int(gvhsize*vscale),int(gvvsize*vscale)))
		gvimg.save(gvfile)
		
		# Updated vertical size - we need this for the offset when combining	
		gvvsize = int(gvimg.size[1])
		
		# Add some text to the image (not the file)
		draw = ImageDraw.Draw(gvimg)
		#draw.text((10, 10),"Detection map",(0,0,0),font=font)
		
		
		# Now we can combine the three images for easier inspection
		spectrafile = dname+'/Spectra/'+outname
		outfile = dname+'/All/'+str(name)+'.png'
		
		# Open the spectra and get its size. Use floats as we need to find the aspect ratio.
		specimg = PIL.Image.open(spectrafile)
		shsize = float(specimg.size[0])
		svsize = float(specimg.size[1])
				
		# We want to resize the spectra to have the same height as the SDSS image,
		# preserving its aspect ratio
		vscale = rvsize/svsize
		
		# Resize the spectral image and store it as a new image
		modspec = specimg.resize((int(shsize*vscale),int(svsize*vscale)),PIL.Image.ANTIALIAS)
		modhsze = modspec.size[0]
		
		
		# Create a new image. Set the horizontal size to be a combination of the three images and the 
		# vertical size to be the RGB height (though the spectra image will now be the same height)
		newim = PIL.Image.new('RGB', (int(modhsze)+int(rhsize)+int(fvhsize)+int(phsize)+15, int(rvsize)), color='white')
		
		# Add the modified spectral image to the new image
		newim.paste(modspec,box=(0,0))
		# Add the SDSS image with a horizotal offset
		newim.paste(rgbimg, box=(int(modhsze)+5,0))
		# Add the flux image with a horizotal offset
		newim.paste(flximg, box=(int(modhsze)+int(rhsize)+10,0))
		# Add the PV image with a horizotal offset
		newim.paste(pvimg, box=(int(modhsze)+int(rhsize)+int(fvhsize)+15,0))
		# Add the detection map image with a horizontal and vertical offset
		newim.paste(gvimg, box=(int(modhsze)+int(rhsize)+10,gvvsize))
		
		newim.save(outfile, 'PNG')
		
		
		# Since PIL seems to have some feckin' stupid problems adding text on Windows, use matplotlib to open
		# the image and annotate it.	
		img = mpimg.imread(outfile)
		fig = plt.figure()
		plt.rc('font', family='serif')
		nx = img.shape[1]
		ny = img.shape[0]
		fig.set_size_inches(float(nx)/100.0,float(ny)/100.0)
		
		# Trick to remove axes and whitespace
		ax = plt.Axes(fig, [0.,0.,1.,1.,])
		ax.set_axis_off()
		fig.add_axes(ax)
		
		# Add labels. This is done to the combined image so the whole thing gets saved as a single image
		# file.
		# Label the SDSS image. Coordinates will be at across, down from top left
		ax.imshow(img)
		hoffset = int(shsize*(rvsize/svsize)) + 25				# Rescaled size of spectrum plus a bit more
		ax.text(hoffset, 25, 'SDSS RGB', fontsize=16, color='red')
		
		# Label the moment map
		hoffset = int(shsize*(rvsize/svsize)) + int(rhsize) + 25
		ax.text(hoffset, 25, 'RA-Dec', fontsize=16, color='red')	
		
		# Label the PV map	
		hoffset = int(shsize*(rvsize/svsize)) + int(rhsize) + int(fhsize*(rvsize/fvsize)*0.5) + 25 + 15
		ax.text(hoffset, 25, 'RA-Vel', fontsize=16, color='red')	
		
		# Label the detection map
		hoffset = int(shsize*(rvsize/svsize)) + int(rhsize) + 25
		voffset = int(fvsize*(rvsize/fvsize)*0.5) + 25	# Accounts for rescaled vertical size of flux map
		ax.text(hoffset, voffset, 'Detection map', fontsize=16, color='black')
		
		plt.savefig(outfile,dpi=100,facecolor='black')
		plt.close(fig)
		
			
	catfile.close()

	print 'All done !'


# If we're resuming from a standard search, find the line in the main catalogue file we need to
# start from
def skiplines(sourcetype):
	# If we're doing the review for the first time, or re-doing the whole lot, then there's no need
	# to skip any lines - we'll want to start from the beginning
	if (mode == 'Review') or (mode == 'Complete'):
		return 0
	
	# If we're resuming after having aborted a search, we'll need to know how we got last time. Find
	# this by comparing the contents of the directories (i.e. sub-catalogues) and main catalogue file. 
	# Look through the sub-catalogues and find the source at the highest line numer.
	# Checks have already been carried out so that we know the catalogues and directories are
	# correct and in agreement	
	if mode == 'Resume':
		print 'Getting number of lines to skip...'
		nmax = 0	# Number of lines in the catalogue file we need to skip
		
		# OLD, DAFT METHOD
		#for infile in ['AcceptedSources.txt','RejectedSources.txt','UncertainSources.txt']:
		#	nlines = 0	# Number of the lines searched each time within each the catalogue, i.e. for
							# each sub-category
							
		#	searchedfile = open(infile,'r')
		#	searchedfile.readline()
			
		#	ocatfile = open('Catalogue.txt','r')
		#	ocatfile.readline()
			
		#	for line in searchedfile:
		#		name, x, y, z, ra, dec, vel, SN, vw, rms = line.split()
		#		tname = str(name)
		#		for line in ocatfile:
		#			name, x, y, z, ra, dec, vel, SN, vw, rms, gid, gs = line.split()
		#			cname = str(name)
		#			
		#			if cname <> tname:
		#				nlines = nlines + 1
		#				
		#			if cname == tname:
		#				nlines = nlines + 1
		#				# Carry on to the next source in the searched file
		#				break
			
		#	# Once the search of each file is complete, update the number of lines we need to skip
		#	# (if necessary)
		#	if nlines > nmax:
		#		nmax = nlines
			
		#	searchedfile.close()
		#	ocatfile.close()
		
		# NEW, SIMPLER METHOD - add up all the lines in each of the subcatalogue files
		infile = open('AcceptedSources.txt','r')
		nlines1 = len(infile.readlines())
		infile.close()
		
		infile = open('RejectedSources.txt','r')
		nlines2 = len(infile.readlines())
		infile.close()
		
		infile = open('UncertainSources.txt','r')
		nlines3 = len(infile.readlines())
		infile.close()
		
		nmax = nlines1 + nlines2 + nlines3 - 3 # Each catalogue also has a header
		
		# Get the total number of sources
		infile = open('Catalogue.txt','r')
		ncat = len(infile.readlines()) - 1     # Take off the header
		infile.close()
		
		# Alert the user if all lines will be skipped
		if nmax == ncat:
			print "It appears you've already completed the review process. Either"
			print "specify which sources (Accepted/Rejected/Uncertain) you want to"
			print "review from scratch or use review mode to begin again for all" 
			print "sources."
			
		if nmax > ncat:
			print 'ERRRK ! Something has gone horribly wrong !'
			
		return nmax



def makecatalogues():
	global goodsources
	global maybesources
	global badsources
	# Create the sub-category catalogue files
	goodfile = open('AcceptedSources.txt','w')
	goodfile.write('#name x y z ra dec vel sn vw rms'+'\n')
	
	badfile  = open('RejectedSources.txt','w')
	badfile.write('#name x y z ra dec vel sn vw rms'+'\n')
	
	maybefile= open('UncertainSources.txt','w')
	maybefile.write('#name x y z ra dec vel sn vw rms'+'\n')

	for thing in goodsources:
		goodfile.write(str(thing)+'\n')
	for thing in badsources:
		badfile.write(str(thing)+'\n')
	for thing in maybesources:
		maybefile.write(str(thing)+'\n')
	
	goodfile.close() 
	badfile.close()
	maybefile.close()



# Opens the spectra images for viewing and lets the user assign their status interactively.
def decisiontime():
	print 'Inspecting images...'
	
	# Default is to assume we're reviewing from scratch
	catfilename = 'Catalogue.txt'
	
	if mode == 'Review':
		if stype == 'All':
			catfilename = 'Catalogue.txt'
		if stype == 'Accepted':
			catfilename = 'AcceptedSources.txt'
		if stype == 'Uncertain':
			catfilename = 'UncertainSources.txt'
		if stype == 'Rejected':
			catfilename = 'RejectedSources.txt'
	
	
	# Get no. sources
	infile = open(catfilename,'r')
	ncands = len(infile.readlines()) - 1
	infile.close()

	# Load the main catalogue file
	catfile = open(catfilename,'r')
	catfile.readline()
	
		
	# Sets the screen resolution, i.e. size of the displayed image files. Unfortunately since we're showing
	# them on screen this really does need to be the actual DPI of the monitor in order to show them at the
	# correct size. Currently this must be entered manually.
	dpi = 80.0
	
	# Create arrays to hold the source data. If we're starting the review from scratch, these can
	# be blank. If we're resuming, we'll have to load them in from files (not yet implemented).ned
	global goodsources
	global maybesources
	global badsources
	goodsources = numpy.asarray([])
	badsources  = numpy.asarray([])
	maybesources= numpy.asarray([])
	 
	
	# If we're resuming the review process, here we need to load in the existing arrays instead of creating 
	# new ones
	if mode == 'Resume' or (mode == 'Review' and stype <> 'All'):
		goodfile = open('AcceptedSources.txt','r')
		goodfile.readline()
		for line in goodfile:
			line = line.splitlines()[0]	# Avoid reading in line breaks
			goodsources = numpy.append(goodsources,[line])
		goodfile.close()
		
		badfile = open('RejectedSources.txt','r')
		badfile.readline()
		for line in badfile:
			line = line.splitlines()[0]
			badsources = numpy.append(badsources,[line])
		badfile.close()
		
		maybefile = open('UncertainSources.txt','r')
		maybefile.readline()
		for line in maybefile:
			line = line.splitlines()[0]
			maybesources = numpy.append(maybesources,[line])
		maybefile.close()


	# Defines what the buttons do, i.e. copy or move images to directories
	# Uses copy2 as this doesn't need the new file name (uses existing name)
	# If we're doing all sources, copy files. This can always be done from the 'All' directory and
	# will over-write any files present/
	# If doing sub-categories, need to move the files.
	# If we're copying files we can just use the supplied path, but if we're moving them, we
	# have to create the file path based on the category we're searching.
	
	class Index:
		
		def accept(self, event):
			global goodsources
			global badsources
			global maybesources
			print name,'accepted !'
			
			# Need to account for Complete mode as well.
			
			if (mode == 'Review' and stype == 'All') or (mode == 'Resume') or (mode == 'Complete'):
				shutil.copy2(filename, dname+'/Accepted/')
			if mode == 'Review' and stype <> 'All' :
				filepath = dname+'/'+stype+'/'+str(name)+'.png'
				shutil.move(filepath, dname+'/Accepted/'+str(name)+'.png')
							
			# It's possible that this source is already in the correct list, in which case we do not need
			# to alter it
			sourcepresent = False
			for item in goodsources:
				if str(name) in item:
					sourcepresent = True
						
			# If and only if the source is not found in this array, append it
			if sourcepresent == False:
				goodsources = numpy.append(goodsources, [str(name)+' '+str(x)+' '+str(y)+' '+str(z)+' '+str(ra)+' '+str(dec)+' '+str(vel)+' '+str(SN)+' '+str(vw)+' '+str(rms)])
			
			# If the source was found then there's no need to alter this array at this stage. Source will be
			# removed from the other arrays below.			
			
			# If we are redoing a category, we'll need to check the other catalogues and remove this source 
			# from the other working lists if present
			if mode == 'Review' and stype <> 'All':
				# Remove from bad sources array
				j = 0
				for item in badsources:
					if str(name) in item:
						badsources = numpy.delete(badsources, [j], axis=0)
					j = j + 1
				
				# Remove from uncertain sources array
				j = 0
				for item in maybesources:
					if str(name) in item:
						maybesources = numpy.delete(maybesources, [j], axis=0)
					j = j + 1
					
			plt.close()
			
		  
		def maybe(self,event):
			global goodsources
			global badsources
			global maybesources
			print name,'uncertain'
			
			if (mode == 'Review' and stype == 'All') or (mode == 'Resume') or (mode == 'Complete'):
				shutil.copy2(filename, dname+'/Uncertain/')
			if mode == 'Review' and stype <> 'All' :
				filepath = dname+'/'+stype+'/'+str(name)+'.png'
				shutil.move(filepath, dname+'/Uncertain/'+str(name)+'.png')	

			sourcepresent = False
			for item in maybesources:
				if str(name) in item:
					sourcepresent = True
						
			# If the source is not found in this array, append it
			if sourcepresent == False:
				maybesources = numpy.append(maybesources, [str(name)+' '+str(x)+' '+str(y)+' '+str(z)+' '+str(ra)+' '+str(dec)+' '+str(vel)+' '+str(SN)+' '+str(vw)+' '+str(rms)])

							
			if mode == 'Review' and stype <> 'All':
				j = 0
				for item in badsources:
					if str(name) in item:
						badsources = numpy.delete(badsources, [j], axis=0)
					j = j + 1
				
				j = 0	
				for item in goodsources:
					if str(name) in item:
						goodsources = numpy.delete(goodsources, [j], axis=0)
					j = j + 1
			
			plt.close()
		
		
		def reject(self,event):
			global goodsources
			global badsources
			global maybesources
			print name,'REJECTED'
			
			if (mode == 'Review' and stype == 'All') or (mode == 'Resume') or (mode == 'Complete'):
				shutil.copy2(filename, dname+'/Rejected/')
			if mode == 'Review' and stype <> 'All' :
				filepath = dname+'/'+stype+'/'+str(name)+'.png'
				shutil.move(filepath, dname+'/Rejected/'+str(name)+'.png')		

			sourcepresent = False
			for item in badsources:
				if str(name) in item:
					sourcepresent = True
						
			# If the source is not found in this array, append it
			if sourcepresent == False:
				badsources = numpy.append(badsources, [str(name)+' '+str(x)+' '+str(y)+' '+str(z)+' '+str(ra)+' '+str(dec)+' '+str(vel)+' '+str(SN)+' '+str(vw)+' '+str(rms)])


			if mode == 'Review' and stype <> 'All':
				j = 0
				for item in maybesources:
					if str(name) in item:
						maybesources = numpy.delete(maybesources, [j], axis=0)
					j = j + 1
				
				j = 0	
				for item in goodsources:
					if str(name) in item:
						goodsources = numpy.delete(goodsources, [j], axis=0)
					j = j + 1

			plt.close()
		
		
		def queryNED(self,event):
			fullra = ra.split(':')
			fulldc = dec.split(':')
		
			rahh = float(fullra[0])
			ramm = float(fullra[1])
			rass = float(fullra[2])
			# For NED, decimal RA runs from 0-24, not 0-365
			decra= (rahh + (ramm/60.0) + (rass/3600.0))#*15.0
		
			dcdd = float(fulldc[0])
			dcmm = float(fulldc[1])
			dcss = float(fulldc[2])
			decdc= dcdd + (dcmm/60.0) + (dcss/3600.0)
			
			# Strictly speaking we should use the beam radius rather than diameter, but there's no position-
			# matching so use diameter instead
			beamsearch = beams*60.0
						
			url='http://ned.ipac.caltech.edu/cgi-bin/objsearch?in_csys=Equatorial&in_equinox=J2000.0&lon='+str(decra)+str('&lat=')+str(decdc)+str('&radius=')+str(beamsearch)+'1.75&search_type=Near+Position+Search'
			webbrowser.open(url,new=1)
			
		
		def queryNEDstrict(self,event):
			fullra = ra.split(':')
			fulldc = dec.split(':')
		
			rahh = float(fullra[0])
			ramm = float(fullra[1])
			rass = float(fullra[2])
			# For NED, decimal RA runs from 0-24, not 0-365
			decra= (rahh + (ramm/60.0) + (rass/3600.0))#*15.0
		
			dcdd = float(fulldc[0])
			dcmm = float(fulldc[1])
			dcss = float(fulldc[2])
			decdc= dcdd + (dcmm/60.0) + (dcss/3600.0)
			
			# Strictly speaking we should use the beam radius rather than diameter, but there's no position-
			# matching so use diameter instead
			beamsearch = beams*60.0
		
			url='http://ned.ipac.caltech.edu/cgi-bin/objsearch?in_csys=Equatorial&in_equinox=J2000.0&lon='+str(decra)+'&lat='+str(decdc)+'&radius='+str(beamsearch)+'&hconst=73&omegam=0.27&omegav=0.73&corr_z=1&z_constraint=Available&z_value1=&z_value2=&z_unit=z&search_type=Near+Position+Search&ot_include=ANY&in_objtypes1=Galaxies&in_objtypes1=GPairs&in_objtypes1=GTriples&in_objtypes1=GGroups&in_objtypes1=GClusters&nmp_op=ANY&out_csys=Equatorial&out_equinox=J2000.0&obj_sort=Distance+to+search+center&of=pre_text&zv_breaker=30000.0&list_limit=5&img_stamp=YES'
			
			webbrowser.open(url,new=1)
			
			
		def querySDSS(self, event):
			fullra = ra.split(':')
			fulldc = dec.split(':')
		
			rahh = float(fullra[0])
			ramm = float(fullra[1])
			rass = float(fullra[2])
			
			decra= (rahh + (ramm/60.0) + (rass/3600.0))*15.0
		
			dcdd = float(fulldc[0])
			dcmm = float(fulldc[1])
			dcss = float(fulldc[2])
			decdc= dcdd + (dcmm/60.0) + (dcss/3600.0)
			
			# Set the image size to be equal to the beam size at the SDSS maximum resolution
			imres = 0.4		# SDSS smallest pixel scale in arcseconds
			imsize = int(beams/(imres/3600.0))
		
			# If the image size exceeds the SDSS maximum size, alter the resolution
			if imsize > 2048:
				imsize = 2048
				imres = (beams*3600.0) / float(imsize)	# Pixel scale in arcseconds

			url = 'http://skyserver.sdss.org/dr13/en/tools/chart/image.aspx?ra='+str(decra)+str('&dec=')+str(decdc)+'&scale=0.4&width='+str(imsize)+'&height='+str(imsize)	
			webbrowser.open(url,new=1)

			
		def abort(self,event):
			print 'Safely exiting. You may resume at any time !'
			print 'Writing catalogues...'
			makecatalogues()
			plt.close()
			sys.exit()
		

	nexamined = 0
	
	
	# Get the number of lines to skip if resuming
	nskip = skiplines(stype)
	
	# Skip lines. Note that we have already skipped the header line.
	for i in range(nskip):
		catfile.readline()
	
	print 'Number of lines to skip from main catalogue :',nskip
	
	print 'Examining sources in file',catfilename
	
	for line in catfile:
		if catfilename == 'Catalogue.txt':
			name, x, y, z, ra, dec, vel, SN, vw, rms, gid, gs = line.split()
		else:
			name, x, y, z, ra, dec, vel, SN, vw, rms = line.split()
		
		nexamined = nexamined + 1
		print 'Doing source',nexamined,'of',ncands-nskip
	
		readdir = dname+'/All/'
	
		# Setup directories to hold the sources
		if not os.path.exists('Accepted'):
			os.mkdir('Accepted')
		
		if not os.path.exists('Uncertain'):
			os.mkdir('Uncertain')

		if not os.path.exists('Rejected'):
			os.mkdir('Rejected')
	
	
		filename = readdir+str(name)+'.png'
	
		im_data = plt.imread(filename)
		height, width, depth = im_data.shape
			
		figsize = width / float(dpi), height / float(dpi)
	
		fig = plt.figure(figsize=figsize)
		fig.canvas.set_window_title('Source review')
		win = plt.gcf().canvas.manager.window
		
		def disable_event():
			pass
		
		win.protocol("WM_DELETE_WINDOW", disable_event)
		
		#win.overrideredirect(1)
		
		ax = fig.add_axes([0, 0, 1, 1])
		
	
		ax.axis('off')
	
		callback = Index()
	
		# CREATE BUTTONS
		# Format : Left position, vertical position, width, height
		axaccpt = plt.axes([0.84, 0.43, 0.04, 0.05])
		baccpt = Button(axaccpt, 'Accept')
		baccpt.on_clicked(callback.accept)
	
		axmaybe = plt.axes([0.8975, 0.43, 0.04, 0.05])
		bmaybe  = Button(axmaybe, 'Maybe')
		bmaybe.on_clicked(callback.maybe)
	
		axrejct = plt.axes([0.955, 0.43, 0.04, 0.05])
		brejct  = Button(axrejct, 'Reject')
		brejct.on_clicked(callback.reject)
		
		axned   = plt.axes([0.88, 0.27,  0.075,0.05])
		bqned   = Button(axned, 'Query NED (all)')
		bqned.on_clicked(callback.queryNED)
		
		axneds   = plt.axes([0.88, 0.21,  0.075,0.05])
		bqneds   = Button(axneds, 'Q. NED (galaxies)')
		bqneds.on_clicked(callback.queryNEDstrict)
		
		axsdss  = plt.axes([0.88, 0.15,  0.075,0.05])
		bqsdss   = Button(axsdss, 'Query SDSS')
		bqsdss.on_clicked(callback.querySDSS)
		
		axabort = plt.axes([0.88, 0.05, 0.075, 0.05])
		babort  = Button(axabort, 'Save & Exit')
		babort.on_clicked(callback.abort)
	
	
		ax.imshow(im_data, cmap='gray', aspect='equal')
	
		plt.show()

	catfile.close()
	
	
	# Now that the review is complete, we can write the separate catalogue files using the arrays.
	# Can safely over-write these if we're resuming as the arrays will contain the whole set of
	# sources for each category.
	print 'Writing catalogues...'
	makecatalogues()
	



# MAIN CODE
# First do all the cleaning to remove previous files and directories, as necessary
# If we're starting from scratch, remove everything
print 'Cleaning up old files and directories if necessary...'
if mode == 'Search' or mode == 'Complete':
	# Remove all existing catalogues and spectra
	cleansourcedirs()			# Removes the master directories holding all spectra images
	cleansubdirs()				# Removes the sub-directories containing the sub-categorised spectra
	cleanimagedir()			# Removes the directory holding the SDSS /Flux maps images
	cleanbasiccatalogues()	# Removes catalogues files of accepted/rejected/uncertain sources 
	cleansubcatalogues()		# Removes the catalogue files for all sources

# If the source-finding was completed but the review process not started, or we want to re-review all
# sources, then only remove the sub-catageory catalogues and directories
if mode == 'Review' and stype == 'All':
	cleansubdirs()
	cleansubcatalogues()	

# If we want to resume the reviewing process, first check that the current files and directories
# list the correct sources. If they don't, we'll likely have to start over. Program will halt
# if any disagreeemtns found.
if mode == 'Resume':
	print 'Checking existing catalogues match directory contents...'
	sourcecheck()



if mode == 'Search' or mode == 'Complete':
	# Main file which will hold every candidate detected
	candfile = open('FullSearchResults.txt','w')
	candfile.write('#name x y z ra dec vel sn vw rms'+'\n')
	
	ncandidates = 0

	for xp in range(0,sizex):	# Use range of 100 for testing, otherwise sizex
		print xp
		for yp in range(0,sizey):			
			# Check the average polarisation first. If it passes, then do the individual polarisations.
			avspectra = extractspec(xp,yp,zmin,zmax,style,'Average')
			# Fit and subtract the baseline and add a S/N column
			avspectra,avrms = spectralfit(avspectra, dosub=True, fitsn=True)
			avpeaksn, avpeakchan = searchspec(avspectra,xp,yp,avrms)  # Modify this routine to return variables
			if avpeaksn > SNt:
				avvw = vwidth(avspectra,avpeakchan)
				avvel = velchan(avspectra,avpeakchan)
				# In single polarisation mode, we can simply write the candidate to a catlogue if it meets the
				# detection criteria
				if dupol == 0:
					if avvw > VWt:
						ncandidates = ncandidates + 1
						candname = 'GLADOS_'+str(ncandidates).zfill(2)
						ra, dec = wcspix(xp,yp)
						candfile.write(str(candname)+' '+str(xp)+' '+str(yp)+' '+str(avpeakchan)+' '+str(ra)+' '+str(dec)+' '+str(avvel)+' '+str(avpeaksn)+' '+str(avvw)+' '+str(avrms)+'\n')
						
									
				if dupol == 1:
					spectraA = extractspec(xp,yp,zmin,zmax,style,'A')
					spectraA,rmsA = spectralfit(spectraA, dosub=True, fitsn=True)
					peaksnA, peakchanA = searchspec(spectraA,xp,yp,rmsA)
					if peaksnA > SNtp:
						vwA = vwidth(spectraA,peakchanA)
						velA = velchan(spectraA,peakchanA)
						# If all criteria are satisifed for main cube and A cube, do the same for the B cube
						if (abs(velA - avvel) <= Vtol) and (vwA >= VWtp):
							spectraB = extractspec(xp,yp,zmin,zmax,style,'B')
							spectraB,rmsB = spectralfit(spectraB, dosub=True, fitsn=True)
							peaksnB, peakchanB = searchspec(spectraB,xp,yp,rmsB)
							if peaksnB > SNtp:
								vwB = vwidth(spectraB,peakchanB)
								velB = velchan(spectraA,peakchanB)
								if (abs(velB - avvel) <= Vtol) and (vwB >= VWtp):
									ncandidates = ncandidates + 1
									candname = 'GLADOS_'+str(ncandidates).zfill(2)
									ra, dec = wcspix(xp,yp)
									candfile.write(str(candname)+' '+str(xp)+' '+str(yp)+' '+str(avpeakchan)+' '+str(ra)+' '+str(dec)+' '+str(avvel)+' '+str(avpeaksn)+' '+str(avvw)+' '+str(avrms)+'\n')
		

	print 'Total number of candidates :',ncandidates	

	candfile.close()
	
	# Merge the pixels
	CHELL()
	
	# If the catalogue was produced, make the plots
	if os.path.exists(dname+'/Catalogue.txt') == True:
		WHEATLEY()
	if os.path.exists(dname+'/Catalogue.txt') == False:
		print 'Catalogue file not found, nothing to plot.'
	
if mode == 'Complete' or mode == 'Review' or mode == 'Resume':
	decisiontime()
	


	
# TO DO 
# Would be better to have a consistent naming scheme in both full results and catalogue file
# Needs some way for user to enter starting value, or to check numbers in existing catalogues, or to rename
# all sources once search complete and merging done
# Possibly include naming based on field/velocity
# Would be nice to be able to manually specify mask ranges
# Have DPI found automatically !
